/*==============================================================================
 Fractal XDoclet Plugin - Copyright (C) 2002-2006 INRIA Futurs / LIFL
 Fractal Component Model (contact: fractal@objectweb.org)

 This library is free software; you can redistribute it and/or modify it under 
 the terms of the GNU Lesser General Public License as published by the Free 
 Software Foundation; either version 2.1 of the License, or any later version.

 This library is distributed in the hope that it will be useful, but WITHOUT ANY 
 WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License along 
 with this library; if not, write to the Free Software Foundation, Inc., 
 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

 Initial developer(s): Romain Rouvoy (romain.rouvoy@lifl.fr)
 Contributor(s): .
 --------------------------------------------------------------------------------
 $Id: FractalPlugin.java,v 1.4 2006/03/17 12:39:24 rouvoy Exp $
 ==============================================================================*/
package org.objectweb.fractal.fraclet;

import java.util.ArrayList;
import java.util.Arrays;

import org.generama.JellyTemplateEngine;
import org.generama.QDoxCapableMetadataProvider;
import org.generama.VelocityTemplateEngine;
import org.generama.WriterMapper;
import org.generama.defaults.QDoxPlugin;
import org.objectweb.fractal.fraclet.doclets.FractalAcTag;
import org.objectweb.fractal.fraclet.doclets.FractalAcTagImpl;
import org.objectweb.fractal.fraclet.doclets.FractalBcTag;
import org.objectweb.fractal.fraclet.doclets.FractalBcTagImpl;
import org.objectweb.fractal.fraclet.doclets.FractalItfTagImpl;
import org.objectweb.fractal.fraclet.doclets.FractalLogTag;
import org.objectweb.fractal.fraclet.doclets.FractalLogTagImpl;
import org.objectweb.fractal.fraclet.doclets.FractalRcTagImpl;
import org.objectweb.fractal.fraclet.doclets.TagLibrary;

import com.thoughtworks.qdox.model.JavaClass;
import com.thoughtworks.qdox.model.JavaField;

/**
 * Abstract Fractal plugin supporting Jelly and Velocity as template engines.
 * @author <a href="mailto:Romain.Rouvoy@lifl.fr">Romain Rouvoy</a>
 * @version $Revision: 1.4 $
 */
public abstract class FractalPlugin extends QDoxPlugin {
    /** Fractal control interface package exluded by the generation engine. */
    protected static final String FC_CONTROL = "org.objectweb.fractal.api.control";

    protected static final String ATTR_NAME = "AttributeController";

    protected static final String ATTR_CTRL = FC_CONTROL + "." + ATTR_NAME;

    protected static final String OBJECT = "java.lang.Object";

    /*----------------------------------------------------------------------------*/
    /* CONSTRUCTORS */
    /*----------------------------------------------------------------------------*/
    /**
     * Constructor used to generate XML files.
     * @param jelly Jelly template engine used to generate artifacts.
     * @param qdox QDox source code metadata provider.
     * @param writer the output writer.
     */
    public FractalPlugin(JellyTemplateEngine jelly, 
            QDoxCapableMetadataProvider qdox, WriterMapper writer) {
        super(jelly, qdox, writer);
        new TagLibrary(qdox);
    }

    /**
     * Constructor used to generate Java or Property files.
     * @param velocity Velocity template engine used to generate artifacts.
     * @param qdox QDox source code metadata provider.
     * @param writer the output writer.
     */
    public FractalPlugin(VelocityTemplateEngine velocity,
            QDoxCapableMetadataProvider qdox, WriterMapper writer) {
        super(velocity, qdox, writer);
        new TagLibrary(qdox);
    }

    /*----------------------------------------------------------------------------*/
    /* COMPONENT FEATURE */
    /*----------------------------------------------------------------------------*/
    /**
     * Provides the short name of the component depending on the filename.
     * @param cls the java class source file parsed.
     * @return the fully qualified name of the associated definition.
     */
    public String name(final JavaClass cls) {
        int dot = getDestinationFilename(cls).indexOf(".");
        return getDestinationFilename(cls).substring(0, dot);
    }

    /**
     * Provides the fully qualified name of the component depending on the filename.
     * @param cls the java class source file parsed.
     * @return the fully qualified name of the associated definition.
     */
    public String fullname(final JavaClass cls) {
        String prefix = (getDestinationPackage(cls).equals("") ? ""
                : getDestinationPackage(cls) + ".");
        return prefix + name(cls);
    }

    /**
     * Provides the prefix of the package (empty if no package).
     * @param cls the class used.
     * @return the prefix of the package (e.g., org.objectweb.fractal. ).
     */
    public String packagePrefix(final JavaClass cls) {
        return (cls.getPackage() == null ? "" : cls.getPackage() + ".");
    }

    /**
     * Checks if a component glue need to be generated.
     * @param cls the class to check.
     * @return true if some component glue should be generated.
     */
    public boolean hasGlue(final JavaClass cls) {
        return hasBindings(cls, true) || hasAttributes(cls, true)
                || hasReflects(cls, true) || hasLogger(cls, true);
    }

    protected static final String ITF_NAME = FractalItfTagImpl.NAME;
    
    /**
     * Checks if the class defines fractal interface(s).
     * @param cls the class to check.
     * @return true if at least one fractal interface is defined.
     */
    public boolean isInterface(final JavaClass cls) {
        return cls.getTagByName(ITF_NAME, false) != null;
    }

    /**
     * Checks if the class is considered as a component.
     * @param cls the class to check.
     * @return true if cls or its parent has a doclet tag.
     */
    public boolean isComponent(final JavaClass cls) {
        return !(cls.getFullyQualifiedName().equalsIgnoreCase(OBJECT)
                || isController(cls) || cls.isInterface());
    }

    /**
     * Checks if the class is a primitive component.
     * @param cls the class to check.
     * @return true if cls is a non-abstract class.
     */
    public boolean isPrimitive(final JavaClass cls) {
        return !(cls.isInner() || cls.isNative() || cls.isAbstract() || cls
                .isInterface());
    }

    /**
     * Checks if the class is a controller (attribute, binding, lifecycle, etc.).
     * @param cls the class to check.
     * @return true if cls is an interface of the org.objectweb.fractal.control
     *         package.
     */
    public boolean isController(final JavaClass cls) {
        return (cls.isInterface() && packagePrefix(cls).startsWith(FC_CONTROL)) || isAttributeController(cls) ;
    }

    /**
     * Checks if the interface is an attribute controller.
     * @param cls the java class source file parsed.
     * @return true if the java class source file describes an attribute
     *         controller.
     */
    public boolean isAttributeController(final JavaClass cls) {
        return cls.isA(ATTR_CTRL) && cls.isInterface();
    }

    /*----------------------------------------------------------------------------*/
    /* CLASSES HANDLING */
    /*----------------------------------------------------------------------------*/
    /**
     * Provides a list of business classes at the first level.
     * @param cls the class considered.
     * @return the list of business interfaces and class.
     */
    protected JavaClass[] extend(final JavaClass cls) {
        ArrayList list = new ArrayList() ;
        JavaClass[] itfs = cls.getImplementedInterfaces();
        for (int i=0;i<itfs.length;i++)
            if (!isController(itfs[i]))
                list.add(itfs[i]);
        JavaClass sup = cls.getSuperJavaClass();
        if (!sup.getFullyQualifiedName().equals(OBJECT))
            list.add(sup);
        return (JavaClass[]) list.toArray(new JavaClass[list.size()]);
    }
    
    /*----------------------------------------------------------------------------*/
    /* FIELDS HANDLING */
    /*----------------------------------------------------------------------------*/
    /**
     * Provides a list of filtered fields.
     * @param cls the java class source file parsed.
     * @param doclet the name of the doclet to retrieve.
     * @return the list of fields having the doclet tag.
     */
    protected JavaField[] fields(final JavaClass cls, final String doclet,
            final boolean all) {
        ArrayList result = new ArrayList();
        if (all && cls.getSuperJavaClass() != null)
            result.addAll(Arrays.asList(fields(cls.getSuperJavaClass(), doclet,
                    all)));
        JavaField[] fields = cls.getFields();
        for (int i = 0; i < fields.length; i++)
            if (fields[i].getTagByName(doclet) != null)
                result.add(fields[i]);
        return (JavaField[]) result.toArray(new JavaField[0]);
    }

    /*----------------------------------------------------------------------------*/
    /* BINDING FEATURE */
    /*----------------------------------------------------------------------------*/
    protected static final String BC_NAME = FractalBcTagImpl.NAME;
    
    /**
     * Provides the list of bindings defined by the class.
     * @param cls the java class source file parsed.
     * @return the list of fields associated to fractal client interfaces.
     */
    public JavaField[] bindings(final JavaClass cls, final boolean all) {
        return fields(cls, BC_NAME, all);
    }

    /**
     * Checks if the class defines some binding.
     * @param cls the java class source file parsed.
     * @return true if the java class has fractal client interfaces.
     */
    public boolean hasBindings(final JavaClass cls, final boolean all) {
        return bindings(cls, all).length > 0;
    }

    /**
     * Computes the binding name (the field name if no tag specified).
     * @param fld the java field source file parsed.
     * @return the name of the binding.
     */
    public String bindingName(final JavaField fld) {
        FractalBcTag tag = (FractalBcTag) fld.getTagByName(BC_NAME);
        return tag.getName_()!=null?tag.getName_():fld.getName();
    }

    /**
     * Computes the binding signature (the field signature if no tag specified).
     * @param fld the java field source file parsed.
     * @return the signature of the binding.
     */
    public String bindingSignature(final JavaField fld) {
        FractalBcTag tag = (FractalBcTag) fld.getTagByName(BC_NAME);
        return tag.getSignature()!=null?tag.getSignature():fld.getType().getValue();
    }

    /*----------------------------------------------------------------------------*/
    /* ATTRIBUTE FEATURE */
    /*----------------------------------------------------------------------------*/
    protected static final String AC_NAME = FractalAcTagImpl.NAME;
    
    /**
     * Provides the list of component attribute fields.
     * @param cls the java class source file parsed.
     * @return the list of fields associated to fractal attributes.
     */
    public JavaField[] attributes(final JavaClass cls, final boolean all) {
        return fields(cls, AC_NAME, all);
    }

    /**
     * Checks if the component defines some attribute.
     * @param cls the java class source file parsed.
     * @return true if the JavaClass has fractal attributes.
     */
    public boolean hasAttributes(final JavaClass cls, final boolean all) {
        return attributes(cls, all).length > 0;
    }

    /**
     * Return the name of super attribute controller.
     * @param cls the class considered.
     * @return the name of the nearer class needing an attribute controller.
     */
    public String superAttributeController(final JavaClass cls) {
        JavaClass sup = cls.getSuperJavaClass();
        if (sup.getFullyQualifiedName().equals(OBJECT))
            return ATTR_NAME;
        if (hasAttributes(sup, false))
            return sup.getName() + ATTR_NAME;
        return superAttributeController(sup);
    }

    /**
     * Returns the name of the attribute controller.
     * @param cls the class considered.
     * @return the name of the nearer class needing an attribute controller.
     */
    public String attributeController(final JavaClass cls) {
        if (hasAttributes(cls, false))
            return cls.getName() + ATTR_NAME;
        else 
            return hasAttributes(cls, true)?superAttributeController(cls):"";
    }

    /**
     * Computes the name of the attribute (field name if no tag specified).
     * @param fld the java field source file parsed.
     * @return the name of the associated attribute.
     */
    public String attributeName(final JavaField fld) {
        FractalAcTag tag = (FractalAcTag) fld.getTagByName(AC_NAME);
        return tag.getName_()!=null?tag.getName_().toLowerCase():fld.getName().toLowerCase();
    }

    /*----------------------------------------------------------------------------*/
    /* REFLECTING FEATURE */
    /*----------------------------------------------------------------------------*/
    protected static final String RC_NAME = FractalRcTagImpl.NAME;
    
    /**
     * Provides the list of component control fields.
     * @param cls the java class source file parsed.
     * @return the list of fields associated to fractal client interfaces.
     */
    public JavaField[] reflects(final JavaClass cls, final boolean all) {
        return fields(cls, RC_NAME, all);
    }

    /**
     * Checks if the component defines some control fields.
     * @param cls the java class source file parsed.
     * @return true if the java class has fractal client interfaces.
     */
    public boolean hasReflects(final JavaClass cls, final boolean all) {
        return reflects(cls, all).length > 0;
    }

    /*----------------------------------------------------------------------------*/
    /* LOGGING FEATURE */
    /*----------------------------------------------------------------------------*/
    protected static final String LOG_NAME = FractalLogTagImpl.NAME;
    
    /**
     * Provides the reference to the last logger declared.
     * @param cls the reference to the logger.
     * @return the field describing the logger.
     */
    public JavaField logger(final JavaClass cls, final boolean all) {
        JavaField[] fld = fields(cls, LOG_NAME, all);
        return fld.length>0?fld[fld.length - 1]:null;
    }

    /**
     * Checks if the class defines a logger.
     * @param cls the class to check.
     * @return true if a logger is defined in this class (use logger(cls) to
     *         retrieve it).
     */
    public boolean hasLogger(final JavaClass cls, final boolean all) {
        return logger(cls, all) != null;
    }

    /**
     * Provides the name of the logger associated to the field.
     * @param fld the field referencing the logger.
     * @return the name of the logger.
     */
    public String loggerName(final JavaField fld) {
        FractalLogTag tag = (FractalLogTag) fld.getTagByName(LOG_NAME);
        return tag.getName_()!=null?tag.getName_():((JavaClass)fld.getParent()).getName();
    }
}